/*
* Copyright (C) 2001 NetLabs SRL (ejbs@netlabs.com.uy)
*
* This file is part of wccpd.
*
* wccpd is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* wccpd is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with wccpd; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#include "config.h"
/* AIX requires this to be the first thing in the file.  */
#ifndef __GNUC__
# if HAVE_ALLOCA_H
#  include <alloca.h>
# else
#  ifdef _AIX
#pragma alloca
#  else
#   ifndef alloca /* predefined by HP cc +Olibcalls */
char *alloca ();
#   endif
#  endif
# endif
#endif
#include <stdio.h>
#include <stdarg.h>
#include <syslog.h>
#include <unistd.h>
#include <arpa/inet.h>
#include <sys/wait.h>
#include <sys/resource.h>
#include "util.h"


/** Pointer to the shutdown flag. */
static int* fShutdownp = NULL;


/**
* Signal handler to set shutdown flag.
*
* This signal handler just increments the variable pointed by fShutdownp,
* if non-NULL.
*
* @param sig the signal that was triggered.
*/
static RETSIGTYPE
signalHandler (int sig) {
	if (fShutdownp != NULL) {
		(*fShutdownp)++;
	}
#if !defined(HAVE_SIGACTION) && defined(HAVE_SIGNAL)
	signal (sig, signalHandler);
#endif
}

/**
* Gets the string representation of a given IP address.
*
* This function uses a statically allocated buffer, thus each call overwrites
* the previous call's result.
*
* @param ip the IP address to convert, in Network Byte Order.
* @return the quad-dot notation string of the given ip.
*/
char*
ip2str (u_int32 ip) {
	struct in_addr inaddr;

	inaddr.s_addr = ip;
	return inet_ntoa (inaddr);
}

/**
* Executes a system command with the given parameters.
*
* This function is just a wrapper arround fork/execv that forks and executes
* the given command.  The full path to the command is <libexecdir>/wccpd/cmd. 
* It will wait for the command to complete before it returns.
*
* The variable length char* parameter list must be NULL terminated, and must
* not exceed MAX_CMDEXEC_ARGS parameters (not including neither the command
* name nor the trailing NULL).
*
* @param cmd the command to execute.
*/
void
executeCmd (char* cmd, ...) {
	int status = 0;
	char* path = (char*)alloca(sizeof(XSTR(SCRIPTSDIR)) + strlen (cmd) + 1);
	va_list ap;
	int i = 0;
	char* argv [1 + MAX_CMDEXEC_ARGS + 1];

	if (path == NULL) {
		syslog (LOG_ALERT, "executeCmd: memory exausted!");
		return;
	}
	strcat (strcpy (path, XSTR(SCRIPTSDIR) "/"), cmd);

	switch (fork ()) {
	case -1:
		syslog (LOG_WARNING, "executeCmd: %s: %m", cmd);
		break;
	case  0:
		argv [i++] = cmd;
		va_start (ap, cmd);
		for (; i <= MAX_CMDEXEC_ARGS; i++) {
			if ((argv [i] = va_arg (ap, char*)) == NULL) {
				break;
			}
		}
		va_end (ap);
		argv [i++] = NULL;
		execv (path, argv);
		_exit (127);
	default:
		while (wait (&status) == -1);
		if (status != 0) {
			syslog (LOG_WARNING, "executeCmd: %s: %d", cmd, status);
		}
		break;
	}
}

/**
* Make the current process a daemon.
*
* Forks off and closes all files.
*
* @return 0 on the daemon, pid on the parent (which exists) and -1 on error.
*/
pid_t
daemonize () {
	pid_t pid = fork ();

	switch (pid) {
	case 0: /* child */
		break;
	case -1:
		perror ("fork");
		return -1;
	default:
		return pid;
	}

	{
#if defined(OPEN_MAX)
		int i = OPEN_MAX;
#elif defined(FOPEN_MAX)
		int i = FOPEN_MAX;
#else	/* !defined(OPEN_MAX) && !defined(FOPEN_MAX) */
		int i = 4096;
#endif	/* !defined(OPEN_MAX) && !defined(FOPEN_MAX) */
#ifdef HAVE_GETRLIMIT
		struct rlimit rl;
		if (getrlimit (RLIMIT_NOFILE, &rl) != -1) {
			i = rl.rlim_max;
		}
#endif	/* HAVE_GETRLIMIT */
		while (i--) {
			close (i);
		}
	}
#if defined(HAVE_SETSID)
	setsid ();
#elif defined(HAVE_SETPGRP)
	setpgrp ();
#endif	/* defined(HAVE_SETSID) || defined(HAVE_SETPGRP) */
		
	return 0;
}

/**
* Catches SIGTERM and SIGINT.
*
* Each time one of the signals is received, the integer pointed to by fShutdown
* is incremented.
*
* @param fShutdown a pointer to an integer to be incremented.
*/
void
setSignals (int* fShutdown) {
	fShutdownp = fShutdown;
#if defined(HAVE_SIGACTION)
	{
		struct sigaction sa;

		memset (&sa, 0, sizeof(sa));
		sa.sa_handler = signalHandler;
		sigemptyset (&sa.sa_mask);

		sigaction (SIGTERM, &sa, NULL);
		sigaction (SIGINT, &sa, NULL);
	}
#elif defined(HAVE_SIGNAL)
	signal (SIGTERM, signalHandler);
	signal (SIGINT, signalHandler);
#endif	/* defined(HAVE_SIGNAL) */
}
