/*
* Copyright (C) 2001 NetLabs SRL (ejbs@netlabs.com.uy)
*
* This file is part of wccpd.
*
* wccpd is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* wccpd is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with wccpd; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <stdio.h>
#include <memory.h>
#include <unistd.h>
#include <syslog.h>
#include <sys/socket.h>
#include <sys/select.h>
#include <netinet/in.h>
#include "wccpv1.h"

/** The shutdown flag, incremented when the daemon receives a SIGTERM/SIGINT */
static int fShutdown = 0;

/**
* Create a socket for the WCCP port.
*
* @param port the port number to use.
* @return a socket file descriptor, or -1 if there was an error.
*/
static int
createSocket (int port) {
	struct sockaddr_in addr;
	int                fd = socket (PF_INET, SOCK_DGRAM, 0);
	
	if (fd == -1) {
		perror ("socket");
		return -1;
	}
	
	memset (&addr, 0, sizeof (addr));
	addr.sin_family = AF_INET;
	addr.sin_port   = htons (port);
	if (bind (fd, (struct sockaddr*) &addr, sizeof (addr)) == -1) {
		perror ("bind");
		close (fd);
		return -1;
	}
	
	return fd;
}

/**
* Get a single packet from the socket.
*
* @param fd the socket file descriptor used to receive the packet.
* @param timeout wait at most this much for the next packet.
* @param buf the packet's data.
* @param len the length of the packet's data.
* @param from where to store the client's address.
* @param fromlen a pointer to store the length of the client's address.
*/
static int
getPacket (int fd, int timeout, void* buf, int len,
	   struct sockaddr_in* from, socklen_t* fromlen) {
	struct timeval  tv;
	struct timeval* tvp        = NULL;
	fd_set          readfds;

	if (timeout != -1) {
		tv.tv_sec  = timeout;
		tv.tv_usec = 0;
		tvp        = &tv;
	}

	FD_ZERO (&readfds);
	FD_SET  (fd, &readfds);
	if (select (fd + 1,  &readfds,  NULL, NULL, tvp) != 1) {
		return -1;
	}

	return recvfrom (fd, buf, len, 0, (struct sockaddr*) from, fromlen);
}

/**
* Process a single packet, and send a reply if applicable.
*
* If the len is less than or equals to 0, then there was an error
* receiving the packet, so we should ignore the request.
*
* @param fd the socket file descriptor used to send the reply.
* @param buf the packet's data.
* @param len the length of the packet's data.
* @param client the client's address.
* @param client_len a pointer to the length of the client's address.
*/
static void
processPacket (int fd, void* buf, int len,
		struct sockaddr_in* client, socklen_t* client_len) {
	if ((len > 0) && (*client_len == sizeof(struct sockaddr_in))) {
		struct msghdr* msg = WCCP1Receive (client->sin_addr.s_addr,
						   buf, len);

		if (msg != NULL) {
			msg->msg_name    = client;
			msg->msg_namelen = *client_len;
			sendmsg (fd, msg, 0);
			WCCP1FreeReply (msg);
		}
	}
}

/**
* The main loop of the daemon.
*
* It loops until fShutdown is greater than 0.  Each interaction it times-out
* clients, gets a packet and processes it.
*
* @param fd the socket file descriptor used to send/receive packets.
*/
static void
main_loop (int fd) {
	while (!fShutdown) {
		unsigned char      buf [WCCP1_MAX_PACKET_SIZE];
		struct sockaddr_in client;
		socklen_t          client_len = sizeof(client);

		processPacket (fd, buf, getPacket (fd, WCCP1TimeOutEntries (),
			      			   buf, sizeof(buf),
						   &client, &client_len),
			       &client, &client_len);
	}
}

/**
* The main function.
*
* It parses options and optionally daemonizes the program.  Then it creates
* the socket and calls main_loop.
*
* @param argc the number of arguments in the command line.
* @param argv the arguments in the command line.
*/
int
main (int argc, char** argv) {
	int opt    = 0;
	int daemon = 1;
	int port   = WCCP1_PORT;
	int fd;

	while ((opt = getopt (argc, argv, "?dp:")) != EOF) {
		switch (opt) {
		case 'd':
			daemon = 0;
			break;
		case 'p':
			port = atoi (optarg);
			break;
		default:
			fprintf (stderr, "Usage: %s [-d] [-p port]\n", argv [0]);
			_exit (1);
		}
	}

	if (daemon && daemonize ()) {
		exit (0);
	}
	setSignals (&fShutdown);

	if ((fd = createSocket (port)) != -1) {
		syslog (LOG_NOTICE, "WCCP Server daemon started (pid=%d)",
			getpid ());
		WCCP1Init ();
		main_loop (fd);
		WCCP1ClearAllClients ();
		syslog (LOG_NOTICE, "WCCP Server daemon stopped (pid=%d)",
			getpid ());
		return 0;
	}

	return 2;
}
