/*
* Copyright (C) 2001 NetLabs SRL (ejbs@netlabs.com.uy)
*
* This file is part of wccpd.
*
* wccpd is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* wccpd is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with wccpd; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/**
* Defines the high-level structures for WCCP version 1.
*/

#ifndef __WCCP_V1_H__
#define __WCCP_V1_H__ 1

#include <time.h>
#include <sys/uio.h>
#include "wccpv1_proto.h"

/*
 * Server structures
 */

/**
* Defines the data for a WCCP version 1 client.
*/
typedef struct S_WCCP1_CLIENT {
	/** 1 if the client is usable and reachable. */
	int    			 usable   : 1,
	/** 1 if the client has buckets assigned to it. */
				 assigned : 1;
	/** Time this client expires. */
	time_t			 expire;
	/** WCCP1_HERE_I_AM and WCCP1_I_SEE_YOU Protocol Version.
	* (WCCP1_PROTO_VERSION_MIN <= version <= WCCP1_PROTO_VERSION_MAX)
	* (Net Byte Order)
	*
	* We send our WCCP1_I_SEE_YOU with the same version received in
	* the client's WCCP1_HERE_I_AM message.
	*/
	u_int32     version;
	/** The last Received Id sent. (In Host Byte Order) */
	u_int32			 rid;
	/** Entry's data, as to be sent on the wire. (Network Byte Order) */
	WCCP1_I_SEE_YOU_WC_ENTRY data;
	/** Next client in list. */
	struct S_WCCP1_CLIENT*	 next;
} WCCP1_CLIENT;

/* wccpv1_proto.c */
/**
* Process a received WCCP version 1 packet.
*
* This is the entry point for all WCCP version 1 packet processing.
*
* In case of a reply, this function does not fill in the msghdr.msg_name
* and msghdr.msg_namelen, that will be taken care of by the caller.  The
* caller is also responsible of deallocating the reply by calling
* WCCP1FreeReply.
*
* @param ip the IP address of the client that sent this packet.
* @param packet the packet's data.
* @param len the length of the packet, as received from the socket layer.
* @return a pointer to a msghdr with the reply to send to the client thru sendmsg.
*/
extern struct msghdr* WCCP1Receive (u_int32 ip, void* packet, int len);

/**
* Frees a reply from WCCP1Receive.
*
* @param reply the pointer to the msghdr obtained from WCCP1Receive.
*/
extern void WCCP1FreeReply (struct msghdr* reply);

/* wccpv1_clients.c */
/** Initialize WCCP version 1 client database.  */
extern void WCCP1Init ();

/** Get a WCCP version 1 client by IP address.
*
* @param ip the IP address of the client to get.
* @return the client's data, or NULL if no client was found for the given IP.
*/
extern WCCP1_CLIENT* WCCP1GetClientByIP (u_int32 ip);

/**
* Adds a WCCP version 1 client with the given IP address to the list of clients.
*
* @param ip the IP address of the client to add.
* @return the newly created client's data, or NULL if allocation failed.
*/
extern WCCP1_CLIENT* WCCP1AddClient (u_int32 ip);

/**
* Update the global WCCP version 1 change number.
* According to the specs, the change number must be updated each time a
* usable cache is added, deleted, or changes it's hash information.
*/
extern void WCCP1UpdateChangeNumber ();

/**
* Get the current global WCCP version 1 change number.
*
* @return the current change number.
*/
extern u_int32 WCCP1GetChangeNumber ();

/**
* Gets a count of usable client caches.
* Usable caches are those for which reachability has been tested.
*
* @return the number of usable caches.
*/
extern u_int32 WCCP1GetUsableCacheCount ();

/**
* Gets the data to be sent for each usable client cache.
*
* Usable caches are those for which reachability has been tested.
* It fills the given iov vector with pointers to the data element
* of each client.  It is assumed that enough iov entries were
* provided.  This function will typically be called after
* WCCP1GetUsableCacheCount.
*
* @param iov a iov array where the data must be stored.
*/
extern void WCCP1GetUsableCacheData (struct iovec* iov);

/**
* Update the hash information with the given IPs and hashes.
*
* The caller must ensure that all the given IPs in the list
* belong to valid and usable clients before calling this function.
*
* @param wccount the number of cache entries in this update.
* @param ips a vector of IP addresses for this update.
* @param hashes the new hash for each of the above given IPs.
*/
extern void WCCP1UpdateHashInfo (u_int32 wccount, u_int32* ips,
				 WCCP1_HINFO hashes[]);
/**
* Expire timed-out clints, and get when will the next timeout will be.
*
* @return the number of seconds until the next timeout.
*/
extern int WCCP1TimeOutEntries ();

/**
* Disables all WebCaches and clears the clients database.
*/
extern void WCCP1ClearAllClients ();

#endif /* __WCCP_V1_H__ */
