/*
* Copyright (C) 2001 NetLabs SRL (ejbs@netlabs.com.uy)
*
* This file is part of wccpd.
*
* wccpd is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* wccpd is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with wccpd; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <memory.h>
#include <limits.h>
#include <syslog.h>
#include "wccpv1.h"
#include "util.h"


/** The list of clients. */
static WCCP1_CLIENT* client_list  = NULL;
/**
* Change number, updated every time a usable cache is added or deleted,
* or it's hash information changes.
* */
static u_int32       changeNumber = 1;
/**
* An empty Hash Information to compare. (to choose between changeCache and
* disableCache)
*/
static WCCP1_HINFO emptyHINFO  = { 0, 0, 0, 0, 0, 0, 0, 0 };
/**
* Current number of assigned clients
*/
static int assignedClients = 0;


/**
* Get the index of the given IP address in the given array of IPs.
*
* @param ip the IP address to look for.
* @param wccount the number of IP addresses in the array.
* @param the array of IPs to look into.
* @return the index of the IP address in the array, or -1 if not found.
*/
static int
indexOf (u_int32 ip, u_int32 wccount, u_int32* ips) {
	int count = 0;

	for (; count < wccount; count++) {
		if (ip == *(ips++)) {
			return count;
		}
	}

	return -1;
}

/**
* Convert a WCCP1_HINFO structure to a hexadecimal string.
*
* @param hinfo the hash information to convert.
* @param buf the buffer to store the string representation in.
* @return the hexadecimal string representation of the hash information.
*/
static char*
hinfo2str (WCCP1_HINFO hinfo, char* buf) {
	int i = 0;

	for (; i < (sizeof(WCCP1_HINFO) / sizeof(u_int32)); i++) {
		snprintf (buf + (i * 8), 9, "%08X", hinfo [i]);
	}
	return buf;
}

/**
* Enable the WCCP version 1 WebCache redirection engine.
*
* This function is called just before the first cache is enabled.
*/
static void
enableWCCP () {
	syslog (LOG_INFO, "WCCP enabled");
	executeCmd ("enable-wccp", NULL);
}

/**
* Disable the WCCP version 1 WebCache redirection engine.
*
* This function is called just after the last cache is disabled.
*/
static void
disableWCCP () {
	syslog (LOG_INFO, "WCCP disabled");
	executeCmd ("disable-wccp", NULL);
}

/**
* Enable a WCCP version 1 WebCache.
*
* This function is called when a cache must be enabled (buckets
* assigned to it, and not previously assigned).
*
* @param client the client which must be enabled.
*/
static void
enableCache (WCCP1_CLIENT* client) {
	char nhash [2*sizeof(WCCP1_HINFO) + 1];

	if (assignedClients++ == 0) {
		enableWCCP ();
	}
	syslog (LOG_INFO, "WebCache at %s enabled", ip2str (client->data.ip));
	client->assigned = 1;
	executeCmd ("enable-cache", ip2str (client->data.ip),
		    hinfo2str (client->data.hinfo, nhash), NULL);
}

/**
* Disable a WCCP version 1 WebCache.
*
* This function is called when a cache must be disabled (no buckets
* assigned to it).
*
* @param client the client which must be disabled.
*/
static void
disableCache (WCCP1_CLIENT* client) {
	char ohash [64 + 1];

	syslog (LOG_INFO, "WebCache at %s disabled", ip2str (client->data.ip));
	client->assigned = 0;
	executeCmd ("disable-cache", ip2str (client->data.ip),
		    hinfo2str (client->data.hinfo, ohash), NULL);
	if (--assignedClients == 0) {
		disableWCCP ();
	}
}

/**
* Change WCCP version 1 cache information.
*
* This function is called when a cache's hash information is changed.
* It is only when after an enableCache call, and before the disableCache.
* This function also copies the new hash information into the existing
* client data structure.
*
* @param client the client which change it's hash information.
* @param hash the new hash information for this client.
*/
static void
changeCache (WCCP1_CLIENT* client, WCCP1_HINFO hash) {
	char ohash [64 + 1];
	char nhash [64 + 1];

	syslog (LOG_INFO, "WebCache at %s changed hash information",
		ip2str (client->data.ip));
	executeCmd ("change-cache", ip2str (client->data.ip),
		    hinfo2str (client->data.hinfo, ohash),
		    hinfo2str (hash, nhash), NULL);
	memcpy (client->data.hinfo, hash, sizeof(WCCP1_HINFO));
}


/** Initialize WCCP version 1 client database.  */
void
WCCP1Init () {
	changeNumber = random () % time (NULL);
}

/** Get a WCCP version 1 client by IP address.
*
* @param ip the IP address of the client to get.
* @return the client's data, or NULL if no client was found for the given IP.
*/
WCCP1_CLIENT*
WCCP1GetClientByIP (u_int32 ip) {
	WCCP1_CLIENT* client = client_list;

	for (; client != NULL; client = client->next) {
		if (client->data.ip == ip) {
			return client;
		}
	}

	return NULL;
}

/**
* Adds a WCCP version 1 client with the given IP address to the list of clients.
*
* @param ip the IP address of the client to add.
* @return the newly created client's data, or NULL if allocation failed.
*/
WCCP1_CLIENT*
WCCP1AddClient (u_int32 ip) {
	WCCP1_CLIENT* client = malloc (sizeof(WCCP1_CLIENT));

	memset (client, 0, sizeof(WCCP1_CLIENT));
	client->rid  = random () % time (NULL);
	client->next = client_list;
	client_list  = client;

	return client;
}

/**
* Update the global WCCP version 1 change number.
* According to the specs, the change number must be updated each time a
* usable cache is added, deleted, or changes it's hash information.
*/
void
WCCP1UpdateChangeNumber () {
	changeNumber++;
}

/**
* Get the current global WCCP version 1 change number.
*
* @return the current change number.
*/
u_int32
WCCP1GetChangeNumber () {
	return changeNumber;
}

/**
* Gets a count of usable client caches.
* Usable caches are those for which reachability has been tested.
*
* @return the number of usable caches.
*/
u_int32
WCCP1GetUsableCacheCount () {
	u_int32       count  = 0;
	WCCP1_CLIENT* client = client_list;

	for (; client != NULL; client = client->next) {
		if (client->usable) {
			count++;
		}
	}

	return count;
}

/**
* Gets the data to be sent for each usable client cache.
*
* Usable caches are those for which reachability has been tested.
* It fills the given iov vector with pointers to the data element
* of each client.  It is assumed that enough iov entries were
* provided.  This function will typically be called after
* WCCP1GetUsableCacheCount.
*
* @param iov a iov array where the data must be stored.
*/
void
WCCP1GetUsableCacheData (struct iovec* iov) {
	WCCP1_CLIENT* client = client_list;

	for (; client != NULL; client = client->next) {
		if (client->usable) {
			iov->iov_base = &client->data;
			iov->iov_len  = sizeof(WCCP1_I_SEE_YOU_WC_ENTRY);
			iov++;
		}
	}
}

/**
* Update the hash information with the given IPs and hashes.
*
* The caller must ensure that all the given IPs in the list
* belong to valid and usable clients before calling this function.
*
* @param wccount the number of cache entries in this update.
* @param ips a vector of IP addresses for this update.
* @param hashes the new hash for each of the above given IPs.
*/
void
WCCP1UpdateHashInfo (u_int32 wccount, u_int32* ips, WCCP1_HINFO hashes[]) {
	int           changed = 0;
	WCCP1_CLIENT* client  = client_list;

	for (; client != NULL; client = client->next) {
		int idx = indexOf (client->data.ip, wccount, ips);

		if ((idx == -1) || (memcmp (hashes [idx], emptyHINFO,
					    sizeof(WCCP1_HINFO)) == 0)) {
			if (client->assigned) {
				disableCache (client);
			}
		} else if (!client->assigned) {
			memcpy (client->data.hinfo, hashes [idx],
				sizeof(WCCP1_HINFO));
			client->data.u = 0;
			enableCache (client);
			changed++;
		} else if (memcmp (client->data.hinfo, hashes [idx],
				    sizeof(WCCP1_HINFO))) {
			changeCache (client, hashes [idx]);
			changed++;
		}
	}

	if (changed) {
		WCCP1UpdateChangeNumber ();
	}
}

/**
* Expire timed-out clints, and get when will the next timeout will be.
*
* @return the number of seconds until the next timeout.
*/
int
WCCP1TimeOutEntries () {
	time_t         now     = time (NULL);
	int            mdiff   = INT_MAX;
	int            changed = 0;
	WCCP1_CLIENT** clientp = &client_list;
	WCCP1_CLIENT** next;

	for (; *clientp != NULL; clientp = next) {
		int diff = (*clientp)->expire - now;

		next = &(*clientp)->next;
		if (diff <= 0) {
			/* Client expired */
			WCCP1_CLIENT* client = *clientp;

			*clientp = *next;

			if (client->assigned) {
				disableCache (client);
			}
			if (client->usable) {
				changed++;
			}
			free (client);
		} else if (diff < mdiff) {
			mdiff = diff;
		}
	}

	if (changed) {
		WCCP1UpdateChangeNumber ();
	}

	return (mdiff == INT_MAX)? -1 : mdiff;
}

/**
* Disables all WebCaches and clears the clients database.
*/
void
WCCP1ClearAllClients () {
	int           changed = 0;
	WCCP1_CLIENT* client  = client_list;
	WCCP1_CLIENT* next;

	for (; client != NULL; client = next) {
		next = client->next;

		if (client->assigned) {
			disableCache (client);
		}
		if (client->usable) {
			changed++;
		}
		free (client);
	}

	if (changed) {
		WCCP1UpdateChangeNumber ();
	}
}
