/*
* Copyright (C) 2001 NetLabs SRL (ejbs@netlabs.com.uy)
*
* This file is part of wccpd.
*
* wccpd is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* wccpd is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with wccpd; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <malloc.h>
#include <memory.h>
#include <syslog.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include "wccpv1.h"
#include "util.h"


/**
* Build a I SEE YOU reply packet for the given client.
*
* This function does not fill in the msghdr.msg_name and msghdr.msg_namelen,
* that will be taken care of by the caller.  The caller is also responsible
* of deallocating the reply by calling WCCP1FreeReply.
*
* @param client the client that will receive this packet.
* @return a pointer to a msghdr with the reply to send to the client thru sendmsg.
*/
static struct msghdr*
build_I_SEE_YOU (WCCP1_CLIENT* client) {
	struct msghdr*   msg     = malloc (sizeof(struct msghdr));
	u_int32          wccount = WCCP1GetUsableCacheCount ();
	WCCP1_I_SEE_YOU_MSG* reply;
	
	if (msg == NULL) {
		syslog (LOG_ALERT, "I_SEE_YOU: msghdr: memory exausted: %m");
		return NULL;
	}

	if ((reply = malloc (sizeof(WCCP1_I_SEE_YOU_MSG))) == NULL) {
		syslog (LOG_ALERT, "I_SEE_YOU: reply: memory exausted: %m");
		free (msg);
		return NULL;
	}

	/* Build reply message
	* msg.msg_name will be taken care at send time
	*/
	msg->msg_iovlen     = 1 + wccount;
	msg->msg_iov        = malloc (msg->msg_iovlen * sizeof(struct iovec));
	if (msg->msg_iov == NULL) {
		syslog (LOG_ALERT, "I_SEE_YOU: msg_iov(%d) memory exausted: %m",
			msg->msg_iovlen);
		free (reply);
		free (msg);
		return NULL;
	}
	msg->msg_control    = NULL;
	msg->msg_controllen = 0;
	msg->msg_flags      = 0;

	/* I_SEE_YOU fixed header */
	reply->type               = htonl (WCCP1_I_SEE_YOU);
	reply->version            = client->version;
	reply->chgnbr             = htonl (WCCP1GetChangeNumber ());
	reply->rid                = htonl (client->rid);
	reply->wccount            = htonl (wccount);
	msg->msg_iov [0].iov_base = reply;
	msg->msg_iov [0].iov_len  = sizeof(WCCP1_I_SEE_YOU_MSG);

	WCCP1GetUsableCacheData (&(msg->msg_iov [1]));

	return msg;
}

/**
* Handle a HERE I AM request packet for the given client.
*
* In case of a reply, this function does not fill in the msghdr.msg_name
* and msghdr.msg_namelen, that will be taken care of by the caller.  The
* caller is also responsible of deallocating the reply by calling
* WCCP1FreeReply.
*
* @param ip the IP address of the client that sent this packet.
* @param request the WCCP version 1 HERE I AM packet.
* @return a pointer to a msghdr with the reply to send to the client thru sendmsg.
*/
static struct msghdr*
handle_WCCP1_HERE_I_AM (u_int32 ip, WCCP1_HERE_I_AM_MSG* request) {
 	WCCP1_CLIENT* client;
	unsigned long version = ntohl (request->version);

	if ((version < WCCP1_PROTO_VERSION_MIN) ||
	    (version > WCCP1_PROTO_VERSION_MAX)) {
		syslog (LOG_ERR, "HERE_I_AM: unexpected protocol (%d not in [%d,%d]) from %s",
			ntohl (request->version), WCCP1_PROTO_VERSION_MIN,
			WCCP1_PROTO_VERSION_MAX, ip2str (ip));
		return NULL;
	} else if (ntohl (request->hrev)    != WCCP1_HASH_REVISION) {
		syslog (LOG_ERR, "HERE_I_AM: unexpected hash revision (%d != %d) from %s",
			ntohl (request->hrev), WCCP1_HASH_REVISION,
			ip2str (ip));
		return NULL;
	}

	if ((client = WCCP1GetClientByIP (ip)) == NULL) {
		if ((client = WCCP1AddClient (ip)) == NULL) {
			return NULL;
		}

		client->version       = request->version;
		client->data.ip       = ip;
		client->data.hrev     = request->hrev;
		memcpy (client->data.hinfo, request->hinfo,
			sizeof(WCCP1_HINFO));
		client->data.u        = request->u;
		client->data.reserved = request->reserved;
	} else if (client->rid != ntohl (request->rid)) {
		syslog (LOG_NOTICE,
			"HERE_I_AM: received-id mismatched (ours=%d theirs=%d)",
			client->rid, ntohl (request->rid));
		return NULL;
	} else {
		if (!client->usable) {
			client->usable = 1;
			syslog (LOG_INFO, "HERE_I_AM: WebCache at %s became usable",
				ip2str (ip));
			WCCP1UpdateChangeNumber ();
		}
		if (++client->rid == 0) {
			/* Can't be 0 according to specs */
			client->rid = 1;
		}
	}

	client->expire = time (NULL) + WCCP1_HERE_I_AM_TIMEOUT;

	return build_I_SEE_YOU (client);
}

/**
* Validate that all the IPs are valid, usable caches.
*
* @param ip the IP address of the client that sent this packet.
* @param wccount the number of entries in the following array.
* @param ips the array of IPs to validate.
* @return 0 if successful, -1 otherwise.
*/
static int
validate_IPs (u_int32 ip, u_int32 wccount, u_int32* ips) {
	char str_ip [16];

	strncpy (str_ip, ip2str (ip), sizeof(str_ip));
	for (; wccount > 0; wccount--, ips++) {
		WCCP1_CLIENT* client = WCCP1GetClientByIP (*ips);
		if (client == NULL) {
			syslog (LOG_ERR,
				"ASSIGN_BUCKET: assignment for unknown cache (%s) from %s",
				ip2str (*ips), str_ip);
			return -1;
		} else if (!client->usable) {
			syslog (LOG_ERR,
				"ASSIGN_BUCKET: assignment for unusable cache (%s) from %s",
				ip2str (*ips), str_ip);
			return -1;
		}
	}

	return 0;
}

/**
* Build the hashes for all the caches.
*
* It packs the bucket data into the WebCache's hash information bitmap.
*
* The hashes are indexed identical to the IP address array.
*
* @param ip the IP address of the client that sent this packet.
* @param wccount the number of entries in the following array.
* @param hinfo an array of assigned WebCache indexes for each bucket.
* @param hashes an array of bitmap (hash information) for each assigned WebCache.
* @return 0 if successful, -1 otherwise.
*/
static int
build_hashes (u_int32 ip, u_int32 wccount, WCCP1_ASSIGN_BUCKET_HASHINFO* hinfo,
	      WCCP1_HINFO hashes[]) {
	int     i      = 0;
	u_int8* bucket = hinfo->bucket;

	memset (hashes, 0, WCCP1_MAX_CACHES * sizeof (WCCP1_HINFO));
	for (; i < sizeof(hinfo->bucket); i++) {
		u_int8 j = *(bucket++);

		if (j == 0xFF) {
			/* Unassigned */
			continue;
		}

		if (j >= wccount) {
			syslog (LOG_ERR,
				"ASSIGN_BUCKET: index out of bounds (%d>=%d) from %s",
				j, wccount, ip2str (ip));
			return -1;
		} else {
			hashes [j][i/32] |= 0x80000000UL >> (i % 32);
		}
	}

	return 0;
}

/**
* Handle a ASSIGN BUCKET request packet for the given client.
*
* It validates the IP and that it's a valid and usable client, then checks
* the Received Id is equal to the one sent in our last I SEE YOU.  Then
* it checks the IPs in the list, and then builds and updates the new hashes.
*
* @param ip the IP address of the client that sent this packet.
* @param packet the packet's data.
* @param len the length of the packet, as received from the socket layer.
*/
static void
handle_ASSIGN_BUCKET (u_int32 ip, void* packet, int len) {
	WCCP1_ASSIGN_BUCKET_MSG* request = (WCCP1_ASSIGN_BUCKET_MSG*) packet;
	u_int32*                 ips;
	u_int32                  wccount = ntohl (request->wccount);
 	WCCP1_CLIENT*            client;
	WCCP1_HINFO		 hashes [WCCP1_MAX_CACHES];

	if (len < (sizeof(WCCP1_ASSIGN_BUCKET_MSG) + wccount * sizeof(u_int32) +
		   sizeof(WCCP1_ASSIGN_BUCKET_HASHINFO))) {
		syslog (LOG_ERR,
			"ASSIGN_BUCKET: packet too short (len=%d, wccount=%d) from %s",
			len, wccount, ip2str (ip));
		return;
	}

	if ((client = WCCP1GetClientByIP (ip)) == NULL) {
		syslog (LOG_WARNING,
			"ASSIGN_BUCKET: packet from unknown client %s",
			ip2str (ip));
		return;
	} else if (client->rid != ntohl (request->rid)) {
		syslog (LOG_NOTICE,
			"ASSIGN_BUCKET: received-id mismatched (ours=%d theirs=%d)",
			client->rid, ntohl (request->rid));
		return;
	} else if (!client->usable) {
		syslog (LOG_WARNING,
			"ASSIGN_BUCKET: packet from unusable client %s",
			ip2str (ip));
		return;
	}


	if (wccount > WCCP1_MAX_CACHES) {
		syslog (LOG_ERR, "ASSIGN_BUCKET: maximum number of WebCaches exceeded! (%d)",
			wccount);
		return;
	}
		        
	ips = (u_int32*) ((u_int8*) packet + sizeof(WCCP1_ASSIGN_BUCKET_MSG));
	if (validate_IPs (ip, wccount, ips) ||
	    build_hashes (ip, wccount,
			  (WCCP1_ASSIGN_BUCKET_HASHINFO*) (ips + wccount),
			  hashes)) {
		return;
	}

	WCCP1UpdateHashInfo (wccount, ips, hashes);
}

/**
* Process a received WCCP version 1 packet.
*
* This is the entry point for all WCCP version 1 packet processing.
*
* This function does not fill in the msghdr.msg_name and msghdr.msg_namelen,
* that will be taken care of by the caller.  The caller is also responsible
* of deallocating the reply by calling WCCP1FreeReply.
*
* @param ip the IP address of the client that sent this packet.
* @param packet the packet's data.
* @param len the length of the packet, as received from the socket layer.
* @return a pointer to a msghdr with the reply to send to the client thru sendmsg.
*/
struct msghdr*
WCCP1Receive (u_int32 ip, void* packet, int len) {
	u_int32 type;

	if (len < 4) {
		syslog (LOG_ERR, "packet too short (%d) from %s", len,
			ip2str (ip));
		return NULL;
	}

	switch (type = ntohl (*(u_int32*) packet)) {
	case WCCP1_HERE_I_AM:
		if (len < sizeof(WCCP1_HERE_I_AM_MSG)) {
			syslog (LOG_ERR, "HERE_I_AM packet too short (%d) from %s",
				len, ip2str (ip));
		} else {
			return handle_WCCP1_HERE_I_AM (ip,
					(WCCP1_HERE_I_AM_MSG*) packet);
		}
		break;
	case WCCP1_ASSIGN_BUCKET:
		if (len < (sizeof(WCCP1_ASSIGN_BUCKET_MSG) +
			   sizeof(WCCP1_ASSIGN_BUCKET_HASHINFO))) {
			syslog (LOG_ERR, "ASSIGN_BUCKET packet too short (%d) from %s",
				len, ip2str (ip));
		} else {
			handle_ASSIGN_BUCKET (ip, packet, len);
		}
		break;
	default:
		syslog (LOG_ERR, "unexpected packet type (%d) from %s",
			type, ip2str (ip));
	}

	return NULL;
}

/**
* Frees a reply from WCCP1Receive.
*
* @param reply the pointer to the msghdr obtained from WCCP1Receive.
*/
void
WCCP1FreeReply (struct msghdr* reply) {
	free (reply->msg_iov [0].iov_base);
	free (reply->msg_iov);
	free (reply);
}
