/*
* Copyright (C) 2001 NetLabs SRL (ejbs@netlabs.com.uy)
*
* This file is part of wccpd.
*
* wccpd is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* wccpd is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with wccpd; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/**
* Defines the loe-level structures for WCCP version 1.
* These are basically taken from the spec. (draft-forster-wrec-wccp-v1-00)
*/

#ifndef __WCCP_V1_PROTO_H__
#define __WCCP_V1_PROTO_H__ 1

#include "types.h"


/*
 * Constants
 */

/** The WCCP version 1 client UDP port number. */
#define WCCP1_PORT		2048

/** The WCCP version 1 HERE_I_AM minimum protocol version. */
#define WCCP1_PROTO_VERSION_MIN	3
/** The WCCP version 1 HERE_I_AM maximum protocol version. */
#define WCCP1_PROTO_VERSION_MAX	4
/** The WCCP version 1 HERE_I_AM hash information revision. */
#define WCCP1_HASH_REVISION	0

/** The types of packets of WCCP version 1. */
enum WCCPV1_TYPES {
	/** WCCP version 1 HERE I AM messages received from clients. */
	WCCP1_HERE_I_AM     = 7,
	/** WCCP version 1 I SEE YOU messages sent to clients. */
	WCCP1_I_SEE_YOU     = 8,
	/** WCCP version 1 ASSIGN BUCKET messages received from clients. */
	WCCP1_ASSIGN_BUCKET = 9
};

/** The time between WCCP version 1 client HERE I AM messages. */
#define WCCP1_HERE_I_AM_T	10
/** How much time without HERE I AM messages before expiring the client. */
#define WCCP1_HERE_I_AM_TIMEOUT (3 * WCCP1_HERE_I_AM_T)
/** Maximum number of assignable  caches supported. */
#define WCCP1_MAX_CACHES	32

/** Maximum packet size. */
#define WCCP1_MAX_PACKET_SIZE	(sizeof(WCCP1_ASSIGN_BUCKET_MSG) +	 \
				 (sizeof(u_int32) * WCCP1_MAX_CACHES) + \
				 sizeof(WCCP1_ASSIGN_BUCKET_HASHINFO))


/*
 * Message formats
 *
 * Everything is in Network Byte Order.
 */

/** Hash information type, a 8-element array of u_int32. */
typedef u_int32 WCCP1_HINFO [8];

/** The WCCP version 1 HERE I AM message structure. */
typedef struct {
	/** Type. (WCCP1_HERE_I_AM) */
	u_int32     type;
	/** Protocol Version. (WCCP1_PROTO_VERSION_MIN <= version <= WCCP1_PROTO_VERSION_MAX) */
	u_int32     version;
	/** Hash Revision. (WCCP1_HASH_REVISION) */
	u_int32     hrev;
	/** Hash Information. */
	WCCP1_HINFO hinfo;
#if defined(WORDS_BIGENDIAN)
	/** Reserved. */
	u_int32 u:1,
	/** Hash Information is historical. */
		reserved:31;
#else /* !defined (WORDS_BIGENDIAN) */
	u_int32 reserved:1,
		u:31;
#endif /* !defined (WORDS_BIGENDIAN) */
	/** Received Id. (must match last Received Id in I_SEE_YOU) */
	u_int32 rid;
} WCCP1_HERE_I_AM_MSG;

/**
* The WCCP version 1 I SEE YOU message structure.
*
* This base structure is followed by 0..wccount WCCP1_I_SEE_YOU_WC_ENTRY
* structures.
*/
typedef struct {
	/** Type. (WCCP1_I_SEE_YOU) */
	u_int32 type;
	/** Protocol Version. (WCCP1_PROTO_VERSION_MIN <= version <= WCCP1_PROTO_VERSION_MAX) */
	u_int32 version;
	/** Change Number. (incremented when a usable cache changes) */
	u_int32 chgnbr;
	/** Received Id. */
	u_int32 rid;
	/** Number of WCs entries to follow. */
	u_int32 wccount;
	/* WCCP1_I_SEE_YOU_WC_ENTRY 0..wccount */
} WCCP1_I_SEE_YOU_MSG;

/** The WCCP version 1 I SEE YOU WebCache entry structure. */
typedef struct {
	/** IP Address of the WebCache. */
	u_int32     ip;
	/** Hash Revision. (as received in HERE I AM packet) */
	u_int32     hrev;
	/** Hash Information. (as received in HERE I AM or ASSIGN BUCKET) */
	WCCP1_HINFO hinfo;
#if defined(WORDS_BIGENDIAN)
	/** Reserved. */
	u_int32 u:1,
	/** Hash Information is historical. */
		reserved:31;
#else /* !defined (WORDS_BIGENDIAN) */
	u_int32 reserved:1,
		u:31;
#endif /* !defined (WORDS_BIGENDIAN) */
} WCCP1_I_SEE_YOU_WC_ENTRY;

/**
* The WCCP version 1 ASSIGN BUCKET message structure.
*
* This base structure is followed by 0..wccount IPs, and then a
* WCCP1_ASSIGN_BUCKET_HASHINFO structure.
* structures.
*/
typedef struct {
	/** Type. (WCCP1_ASSIGN_BUCKET) */
	u_int32 type;
	/** Received Id. (must match last Received Id in I_SEE_YOU) */
	u_int32 rid;
	/** Number of WebCache IPs to follow. */
	u_int32 wccount;
	/* Web-Cache n IP Address list (u_int32) */
	/* WCCP1_ASSIGN_BUCKET_HASHINFO */
} WCCP1_ASSIGN_BUCKET_MSG;

/**
* This is the tail of a WCCP version 1 ASSIGN BUCKET message.
*
* Each value in the array is either 0xFF, which means the bucket is
* unassigned, or the IP's index of the assigned cache (from 0..wccount, max 31).
*/
typedef struct {
	/** The WebCache index for this bucket, or 0xFF if unassigned. */
	u_int8 bucket [256];
} WCCP1_ASSIGN_BUCKET_HASHINFO;

#endif /* __WCCP_V1_PROTO_H__ */
